import user, os, sys, exceptions, re

from xml.sax import ContentHandler
from xml.sax import ErrorHandler
from xml.sax import make_parser
from xml.sax.handler import feature_namespaces

#------------------------------------------------------------------------------

class Parser:

    #--------------------------------------------------------------------------

    def __init__( self, inFileName="", outFileName="" ):

        """
        constructor

        @param inFileName: name of the input file to be converted
        @type inFileName: string

        @param outFileName: name of the output file
        @type outFileName: string
        """

        if os.path.exists( inFileName ):
            self.inFileName = inFileName
        else:
            print "*** Error: input file '%s' doesn't exist" % ( inFileName )
            sys.exit(1)
        self.inFile = open( inFileName, "r" )

        self.outFileName = outFileName
        if outFileName != "":
            self.outFile = open( outFileName, "w" )
        else:
            print "*** Error: output file name is missing"
            sys.exit(1)

    #--------------------------------------------------------------------------

    def set2map( self ):

        """
        Convert a 'set' file into a 'map' file.
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            pathID = data[0]
            seqname = data[1]
            name = data[2]
            start = data[3]
            end = data[4][:-1]

            string = "%s\t%s\t%s\t%s\n" % ( seqname, name, start, end )
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def map2set( self ):

        """
        Convert a 'map' file into a 'set' file.
        """

        path=0
        line = self.inFile.readline()

        while True:

            if line == "":
                break

            path+=1
            string = "%d\t%s" % ( path, line )
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def align2map( self ):

        """
        Convert an 'align' file (output from Blaster) into a 'map' file.
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            qryName = data[0]
            qryStart = data[1]
            qryEnd = data[2]
            sbjName = data[3]
            sbjStart = data[4]
            sbjEnd = data[5]
            BLAST_Eval = data[6]
            BLAST_score = data[7]
            percId = data[8][:-1]

            cmd = "repet\t" + qryName + "\t" + qryStart + "\t" + qryEnd + "\n"
            self.outFile.write( cmd )
            cmd = "repet\t" + sbjName + "\t" + sbjStart + "\t" + sbjEnd + "\n"
            self.outFile.write( cmd )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def align2path( self ):

        """
        Convert an 'align' file into a 'path' file.
        """

        line = self.inFile.readline()
        pathID = 1

        while True:

            if line == "":
                break

            string = "%i\t%s" % ( pathID, line )
            self.outFile.write( string )

            line = self.inFile.readline()
            pathID += 1

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def path2align( self ):

        """
        Convert a 'path' file (output from Matcher) into an 'align' file.
        subject coordinates are always direct (start<end)
        query coordinates are updated accordingly
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            path = data[0]
            qryName = data[1]
            qryStart = int(data[2])
            qryEnd = int(data[3])
            sbjName = data[4]
            sbjStart = int(data[5])
            sbjEnd = int(data[6])
            BLAST_Eval = data[7]
            BLAST_score = data[8]
            percId = data[9][:-1]

            # invert subject coordinates if start > end, and update query coordinates accordingly
            if sbjStart > sbjEnd:
                tmpS = sbjStart
                sbjStart = sbjEnd
                sbjEnd = tmpS
                tmpQ = qryStart
                qryStart = qryEnd
                qryEnd = tmpQ

            # compute a new score: match length on query times identity
            if qryStart < qryEnd:
                matchLengthOnQuery = qryEnd - qryStart + 1
            elif qryStart > qryEnd:
                matchLengthOnQuery = qryStart - qryEnd + 1
            newScore = int( matchLengthOnQuery * float(percId) / 100.0 )

            cmd = "%s\t%i\t%i\t%s\t%i\t%i\t%s\t%i\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, BLAST_Eval, newScore, percId )

            self.outFile.write( cmd )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def pathnum2id( self ):

        """
        Adapt the path IDs as the input file is the concatenation of several 'path' files.
        """

        line = self.inFile.readline()

        dID2count = {}
        count = 1

        while True:

            if line == "":
                break

            data = line.split("\t")

            path = data[0]
            qryName = data[1]
            qryStart = int(data[2])
            qryEnd = int(data[3])
            sbjName = data[4]
            sbjStart = int(data[5])
            sbjEnd = int(data[6])
            BLAST_Eval = data[7]
            BLAST_score = data[8]
            percId = data[9][:-1]

            key_id = path + "-" + qryName + "-" + sbjName
            if key_id not in dID2count.keys():
                newPath = count
                count += 1
                dID2count[ key_id ] = newPath
            else:
                newPath = dID2count[ key_id ]

            cmd = "%i\t%s\t%i\t%i\t%s\t%i\t%i\t%s\t%s\t%s\n" % ( newPath, qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, BLAST_Eval, BLAST_score, percId )

            self.outFile.write( cmd )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def tab2align( self ):

        """
        Convert a 'tab' file (output from Matcher) into an 'align' file.
        """

        line = self.inFile.readline()   # skip the first line
        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")
            qryName = data[0]
            qryStart = data[1]
            qryEnd = data[2]
            qryLength = data[3]
            qryLengthPerc = data[4]
            matchLengthPerc = data[5]
            sbjName = data[6]
            sbjStart = data[7]
            sbjEnd = data[8]
            sbjLength = data[9]
            sbjLengthPerc = data[10]
            Evalue = data[11]
            score = data[12]
            identity = data[13]
            path = data[14]

            string = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, Evalue, score, identity )

            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def tabnum2id( self ):

        """
        Adapt the path IDs as the input file is the concatenation of several 'tab' files, and remove the extra header lines.
        """

        line = self.inFile.readline()
        self.outFile.write( line )
        line = self.inFile.readline()

        dID2count = {}
        count = 1

        while True:

            if line == "":
                break

            data = line.split("\t")

            if data[0] != "query.name":
                qryName = data[0]
                qryStart = data[1]
                qryEnd = data[2]
                qryLength = data[3]
                qryLengthPerc = data[4]
                matchLengthPerc = data[5]
                sbjName = data[6]
                sbjStart = data[7]
                sbjEnd = data[8]
                sbjLength = data[9]
                sbjLengthPerc = data[10]
                Evalue = data[11]
                score = data[12]
                identity = data[13]
                path = data[14]

                key_id = path + "-" + qryName + "-" + sbjName
                if key_id not in dID2count.keys():
                    newPath = count
                    count += 1
                    dID2count[ key_id ] = newPath
                else:
                    newPath = dID2count[ key_id ]

                cmd = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%i\n" % ( qryName, qryStart, qryEnd, qryLength, qryLengthPerc, matchLengthPerc, sbjName, sbjStart, sbjEnd, sbjLength, sbjLengthPerc, Evalue, score, identity, newPath )

                self.outFile.write( cmd )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def cut2map( self ):

        """
        Use a 'cut' file (generated from a fasta file) to generate the corresponding 'map' file.
        """

        bsDB = BioseqDB( self.inFileName )

        for bs in bsDB.db:

            data = bs.header.split()
            seq1 = data[0].split(">")[1]
            seq2 = data[1]
            start = data[-1].split("..")[0]
            end = data[-1].split("..")[0]

            cmd = "%s\t%s\t%s\t%s\n" % ( seq1, seq2, start, end )
            self.outFile.write( cmd )

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def pathrange2set( self ):

        """
        Converts a file from the 'path_range' format to the 'set' format, an entry being a subject (TE) matching on a query (chr).
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")
            string = "%s\t%s\t%s\t%s\t%s\n" % ( data[0], data[4], data[1], data[2], data[3] )

            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def blast2align( self ):

        """
        Convert the output from BLAST (tabulated format) into the 'align' format.
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            if line[0] != "#":
                data = line.split("\t")

                qryName = data[0]
                sbjName = data[1]
                percId = data[2]
                qryStart = data[6]
                qryEnd = data[7]
                sbjStart = data[8]
                sbjEnd = data[9]
                Eval = data[10]
                bitScore = data[11][:-1]

                string = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, Eval, bitScore, percId )

                self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def blastclust2map( self ):

        """

        """

        pass

    #--------------------------------------------------------------------------

    def pals2align( self, sameSequences=False ):

        """
        Convert the output from PALS (GFF2 format) into the 'align' format.
        """

        self.outFile.close()
        tmpFile = "%s_tmp%s" % ( self.outFileName, os.getpid() )
        tmpFileHandler = open( tmpFile, "w" )

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            qryName = data[0]
            source = data[1]
            feature = data[2]
            qryStart = data[3]
            qryEnd = data[4]
            score = data[5]
            strand = data[6]
            frame = data[7]
            attributes = data[8][:-1].split()

            sbjName = attributes[1]
            sbjStart = attributes[2]
            sbjEnd = attributes[3][:-1]
            percId = (1 - float(attributes[-1])) * 100.0

            if strand != "+":
                tmp = sbjStart
                sbjStart = sbjEnd
                sbjEnd = tmp

            if sameSequences \
            and "chunk" in qryName and "chunk" in sbjName \
            and min(int(qryStart),int(qryEnd)) == 1 \
            and min(int(sbjStart),int(sbjEnd)) == 1 \
            and percId == 100.0:
                line = self.inFile.readline()
                continue

            if qryStart < qryEnd:
                string = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, "0.0", score, percId )
            else:
                string = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % ( qryName, qryEnd, qryStart, sbjName, sbjEnd, sbjStart, "0.0", score, percId )

            tmpFileHandler.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        tmpFileHandler.close()

        os.system( "sort -k 1,1 -k 4,4 -k 2,2n -k 3,3n -k 5,5n -k 6,6n -k 8,8n %s > %s" % ( tmpFile, self.outFileName ) )
        os.remove( tmpFile )

    #--------------------------------------------------------------------------

    def align2recon( self ):

        """
        Convert the output file from Blaster ('align' format) into the input for Recon (MSP file).
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            qryName = data[0]
            qryStart = data[1]
            qryEnd = data[2]
            sbjName = data[3]
            sbjStart = data[4]
            sbjEnd = data[5]
            Eval = data[6]
            score = data[7]
            identity = data[8][:-1]

            string = score + " " + identity + " " + qryStart + " " + qryEnd + " " +\
                  qryName + " " + sbjStart + " " + sbjEnd + " " + sbjName + "\n"
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def align2piler( self ):

        """
        Convert the output file from Blaster ('align' format) into the input for Piler (GFF2 format).
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split("\t")

            qryName = data[0]
            qryStart = int(data[1])
            qryEnd = int(data[2])
            sbjName = data[3]
            sbjStart = int(data[4])
            sbjEnd = int(data[5])
            Eval = data[6]
            score = data[7]
            identity = data[8][:-1]

            if qryStart < qryEnd:
                if sbjStart < sbjEnd:
                    strand = "+"
                else:
                    strand = "-"
                    tmp = sbjStart
                    sbjStart = sbjEnd
                    sbjEnd = tmp

            elif qryStart > qryEnd:
                tmp = qryStart
                qryStart = qryEnd
                qryEnd = tmp
                if sbjStart < sbjEnd:
                    strand = "-"
                else:
                    strand = "+"
                    tmp = sbjStart
                    sbjStart = sbjEnd
                    sbjEnd = tmp

            percMis = (100.0 - float(identity)) / 100.0

            string = "%s\tblaster2\thit\t%s\t%s \t%s\t%s\t.\tTarget %s %s %s; maxe %.4f\n" % ( qryName, qryStart, qryEnd, score, strand, sbjName, sbjStart, sbjEnd, percMis )
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def recon2map( self ):

        """
        Convert the output file from Recon (summary/eles) into 'map' format.
        """

        line = self.inFile.readline()   # to avoid the 1st line of comments
        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split()
            clusterID = data[0]
            memberID = data[1]
            strand = data[2]
            seqName = data[3]
            seqStart = data[4]
            seqEnd = data[5]

            if strand == "-1":
                tmp = seqStart
                seqStart = seqEnd
                seqEnd = tmp

            string = "ReconCluster%sMb%s\t%s\t%s\t%s\n" % ( clusterID, memberID, seqName, seqStart, seqEnd )
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def piler2map( self ):

        """
        Convert the output file from Piler into 'map' format.
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split()
            seqName = data[0]
            seqStart = data[3]
            seqEnd = data[4]
            strand = data[6]
            clusterID = data[9]
            memberID = data[12]

            if strand == "-":
                tmp = seqStart
                seqStart = seqEnd
                seqEnd = tmp

            string = "PilerCluster%sMb%s\t%s\t%s\t%s\n" % ( clusterID, memberID, seqName, seqStart, seqEnd )
            #string = "Pile%sFam%s\t%s\t%s\t%s\n" % ( memberID, clusterID, seqName, seqStart, seqEnd )  # old style formatting
            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def RMcat2align( self ):

        """
        Convert the output file from RepeatMasker ('cat' format) into the 'align' format.
        subject coordinates are always direct (start<end)
        query coordinates are updated accordingly
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split()
            try:
                test = int(line[0])
            except ValueError:
                break

            scoreSW = data[0]
            percDiv = data[1]
            percId = 100.0 - float(percDiv)
            percDel = data[2]
            percIns = data[3]

            # query coordinates are always direct (start<end)
            qryName = data[4]
            qryStart = int(data[5])
            qryEnd = int(data[6])
            qryAfterMatch = data[7]

            # subject coordinates can be direct (start<end)
            if data[8] != "C":
                sbjName = data[8]
                sbjStart = int(data[9])
                sbjEnd = int(data[10])
                sbjAfterMatch = data[11]

            # or indirect (start>end): in that case, invert subject coordinates and update query coordinates accordingly
            elif data[8] == "C":
                sbjName = data[9]
                sbjAfterMatch = data[10]
                sbjStart = int(data[12])
                sbjEnd = int(data[11])
                tmpQ = qryStart
                qryStart = qryEnd
                qryEnd = tmpQ

            # compute a new score: match length on query times identity
            if qryStart < qryEnd:
                matchLengthOnQuery = qryEnd - qryStart + 1
            elif qryStart > qryEnd:
                matchLengthOnQuery = qryStart - qryEnd + 1
            newScore = int( matchLengthOnQuery * float(percId) / 100.0 )

            string = "%s\t%i\t%i\t%s\t%i\t%i\t0.0\t%i\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, newScore, percId )

            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def RMcat2path( self ):

        """
        Convert the output file from RepeatMasker ('cat' format) into the 'path' format.
        subject coordinates are always direct (start<end)
        query coordinates are updated accordingly
        """

        line = self.inFile.readline()
        count = 1

        while True:

            if line == "":
                break

            data = line.split()
            try:
                test = int(line[0])
            except ValueError:
                break

            scoreSW = data[0]
            percDiv = data[1]
            percId = 100.0 - float(percDiv)
            percDel = data[2]
            percIns = data[3]

            # query coordinates are always direct (start<end)
            qryName = data[4]
            qryStart = int(data[5])
            qryEnd = int(data[6])
            qryAfterMatch = data[7]

            # subject coordinates can be direct (start<end)
            if data[8] != "C":
                sbjName = data[8]
                sbjStart = int(data[9])
                sbjEnd = int(data[10])
                sbjAfterMatch = data[11]

            # or indirect (start>end): in that case, invert subject coordinates and update query coordinates accordingly
            elif data[8] == "C":
                sbjName = data[9]
                sbjAfterMatch = data[10]
                sbjStart = int(data[12])
                sbjEnd = int(data[11])
                tmpQ = qryStart
                qryStart = qryEnd
                qryEnd = tmpQ

            # compute a new score: match length on query times identity
            if qryStart < qryEnd:
                matchLengthOnQuery = qryEnd - qryStart + 1
            elif qryStart > qryEnd:
                matchLengthOnQuery = qryStart - qryEnd + 1
            newScore = int( matchLengthOnQuery * float(percId) / 100.0 )

            string = "%i\t%s\t%i\t%i\t%s\t%i\t%i\t0.0\t%i\t%s\n" % ( count, qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, newScore, percId )

            self.outFile.write( string )

            line = self.inFile.readline()
            count += 1

        self.inFile.close()
        self.outFile.close()

    #--------------------------------------------------------------------------

    def CENmap2align( self ):

        """
        Convert the output file from Censor ('map' format) into the 'align' format.
        subject coordinates are always direct (start<end)
        query coordinates are updated accordingly
        """

        line = self.inFile.readline()

        while True:

            if line == "":
                break

            data = line.split()

            qryName = data[0]
            strand = data[6]
            if strand == "d":   # if match with direct subject
                qryStart = int(data[1])
                qryEnd = int(data[2])
            elif strand == "c":   # if match with complement subject
                qryStart = int(data[2])
                qryEnd = int(data[1])

            sbjName = data[3]
            sbjStart = int(data[4])
            sbjEnd = int(data[5])

            percId = 100 * float(data[7])
            newScore = int( float(data[9]) * float(data[7]) )

            string = "%s\t%i\t%i\t%s\t%i\t%i\t0.0\t%i\t%s\n" % ( qryName, qryStart, qryEnd, sbjName, sbjStart, sbjEnd, newScore, percId )

            self.outFile.write( string )

            line = self.inFile.readline()

        self.inFile.close()
        self.outFile.close()

#------------------------------------------------------------------------------

class TrfParser:

    #--------------------------------------------------------------------------

    def __init__( self, inFileName, TrfOutFormat, outFileName ):

        """
        constructor

        @param inFileName: name of the input file given to TRF
        @type inFileName: string

        @param TrfOutFormat: format of the TRF ouput ('dat' or 'html')
        @type TrfOutFormat: string

        @param outFileName: name of the output file ('set' format)
        @type outFileName: string
        """

        self.inFileName = inFileName
        self.TrfOutFormat = TrfOutFormat
        if outFileName == "":
            outFileName = inFileName + ".SSRtrf.set"
        self.outFile = open( outFileName, "w" )
        self.nbInSeq = 0
        self.nbPatterns = 0
        self.srchConsSeq = re.compile("^[ATGCNatgcn]+$")

    #--------------------------------------------------------------------------

    def dat2set( self, TrfFileName ):

        """
        Convert the results from TRF ('dat' format) into 'set' format.

        @param TrfFileName: name of the 'dat' file (output from TRF)
        @type TrfFileName: string
        """

        inFile = open( TrfFileName, "r" )

        self.nbPatterns = 0

        while 1:
            line = inFile.readline()

            if line == "":
                break

            data = line.split(" ")

            if len(data) > 1 and "Sequence:" in data[0]:
                self.nbInSeq += 1
                seqName = data[1][:-1]

            if len(data) < 14:
                continue

            self.nbPatterns += 1
            consensus = data[13]
            copyNb = int( float(data[3]) + 0.5 )
            start = data[0]
            end = data[1]

            self.outFile.write( "%i\t(%s)%i\t%s\t%s\t%s\n" % ( self.nbPatterns, consensus, copyNb, seqName, start, end ) )

        self.outFile.close()

        return 0

    #--------------------------------------------------------------------------

    # WARNING: this function contains error !!!
    # use the 'dat2set' function instead

    def html2set( self, TrfFileName ):

        """
        Convert the results from TRF ('txt.html' format) into 'set' format.

        @param TrfFileName: name of the 'txt.html' file (output from TRF)
        @type TrfFileName: string
        """

        inFile = open( TrfFileName, "r" )

        while 1:
            line = inFile.readline()

            if line == "":
                break

            if line.find("Sequence:") != -1:
                seqname = line.replace("Sequence: ","").rstrip()
                m = re.match( "^[0-9]+.+\{Cut\}", seqname )
                if m != None:
                    chr=seqname[m.start(0):m.end(0)-5].rstrip()
                    chr=re.sub("^[0-9]+ ","",chr).lstrip()
                    chk_start = int(seqname[m.end(0):].split("..")[0])
                    print "chr=%s, chk_start=%i" % ( chr, chk_start ); sys.stdout.flush()
                else:
                    print "*** Error in pyRepet.parser.TrfParser"
                    return 1

            if line.find("Indices:") != -1:
                start, end = line.split()[1].split("--")
                start = int(start)
                end = int(end)
                print "start=%i, end=%i" % ( start, end ); sys.stdout.flush()

            if line.find("Period size:") !=- 1:
                copynumber = float(line.split()[4])
                print "copynumber=%.3f" % ( copynumber ); sys.stdout.flush()

            if line.find("Consensus pattern") != -1:
                cons = ""
                while 1:
                    line = inFile.readline().rstrip()
                    if line == "" or self.srchConsSeq.match(line) == None:
                        break
                    cons += line
                self.nbPatterns += 1
                astart = start + chk_start - 1
                aend = end + chk_start - 1
                string = "%d\t(%s)%d\t%s\t%s\t%s" % ( self.nbPatterns, cons, int(copynumber+0.5), chr, astart, aend )
                print string; sys.stdout.flush()
                self.outFile.write( "%s\n" % (string) )

        return 0

    #--------------------------------------------------------------------------

    def clean( self ):

        """
        Remove output files from TRF to keep only the 'set' file.

        @param outFormat: 'dat' or 'html'
        @type outFormat: string
        """

        cmd = "rm -f %s.*.%s" % ( self.inFileName, self.TrfOutFormat )
        log = os.system( cmd )
        if log != 0:
            print "*** Error in pyRepet.parser.Parser.TrfParser.clean"
            sys.exit(1)

#------------------------------------------------------------------------------

# used in MrepsParser

class FindRep( ContentHandler ):

    #--------------------------------------------------------------------------

    def __init__(self,fileoutname,filter,count=0):
        self.inWindowContent=0
        self.inSeqNameContent=0
        self.inStartContent=0
        self.inEndContent=0
        self.inPeriodContent=0
        self.inUnitContent=0
        self.inScoreContent=0
        self.count=count
        self.fileout=open(fileoutname,"w")
        self.filter=filter

    #--------------------------------------------------------------------------

    def startElement(self,name,attrs):
        if name=="window":
            self.inWindowContent=1
        elif name=="sequence-name":
            self.inSeqNameContent=1
            self.seqname=""
        elif name=="repeat":
            self.inRepContent=1
            self.start=""
            self.end=""
            self.period=""
            self.type={}
        elif name=="start":
            self.inStartContent=1
        elif name=="end":
            self.inEndContent=1
        elif name=="period":
            self.inPeriodContent=1
        elif name=="unit":
            self.inUnitContent=1
            self.unit=""
        elif name=="score":
            self.inScoreContent=1
            self.score=""

    #--------------------------------------------------------------------------

    def characters(self,ch):
        if self.inSeqNameContent:
            self.seqname+=ch
        elif self.inStartContent:
            self.start+=ch
        elif self.inEndContent:
            self.end+=ch
        elif self.inPeriodContent:
            self.period+=ch            
        elif self.inUnitContent:
            self.unit+=ch            
        elif self.inScoreContent:
            self.score+=ch            

    #--------------------------------------------------------------------------

    def endElement(self,name):
        if name=="window":
            self.inWindowContent=0
        elif name=="sequence-name":
            self.inSeqNameContent=0
        elif name=="repeat":
            self.inRepContent=0
            start=int(self.start)
            end=int(self.end)
            period=int(self.period)
            score=float(self.score)
            if score>self.filter:
                return
            max=0
            self.count+=1
            for k,n in self.type.items():
                if n>max:
                    max=n
                    k_max=k

            m=re.match("^[0-9]+.+\{Cut\}",self.seqname)
            if m!=None:
                seqname=self.seqname[m.start(0):m.end(0)-5].rstrip()
                seqname=re.sub("^[0-9]+ ","",seqname).lstrip()
                tok=self.seqname[m.end(0):].split("..")
                astart=start+int(tok[0])-1
                aend=end+int(tok[0])-1
            else:
                astart=start
                aend=end
                seqname=self.seqname
            if len(k_max) > 100:
                k_max=k_max[:48]+"..."+k_max[-51:]
            strout="%d\t(%s)%d\t%s\t%d\t%d"%\
                               (self.count,k_max,(abs(start-end)+1)/period,\
                                seqname,astart,aend)
            self.fileout.write("%s\n"%(strout))
            print strout

        elif name=="start":
            self.inStartContent=0
        elif name=="end":
            self.inEndContent=0
        elif name=="period":
            self.inPeriodContent=0
        elif name=="score":
            self.inScoreContent=0
        elif name=="unit":
            self.inUnitContent=0
            if self.type.has_key(self.unit):
                self.type[self.unit]+=1
            else:
                self.type[self.unit]=1

#------------------------------------------------------------------------------

class MrepsParser:

    #--------------------------------------------------------------------------

    def __init__( self, inFileName, MrepsOutFileName, outFileName, errorFilter ):

        """
        constructor

        @param inFileName: name of the input file given to Mreps
        @type inFileName: string

        @param MrepsOutFileName: name of the output file from Mreps ('xml' format)
        @type MrepsOutFileName: string

        @param outFileName: name of the output file ('set' format)
        @type outFileName: string

        @param errorFilter: error filter
        @type errorFilter: float
        """

        self.inFileName = inFileName
        self.MrepsOutFileName = MrepsOutFileName
        if outFileName == "":
            outFileName = inFileName + ".SSRtrf.set"
        self.outFileName = outFileName
        self.errorFilter = errorFilter

    #--------------------------------------------------------------------------

    def xml2set( self ):

        xmlParser = make_parser()
        xmlParser.setFeature( feature_namespaces, 0 )
        xmlParser.setContentHandler( FindRep( self.outFileName, self.errorFilter, 0 ) )

        MrepsOutFile = open( self.MrepsOutFileName, "r" )
        xmlParser.parse( MrepsOutFile )
        MrepsOutFile.close()

    #--------------------------------------------------------------------------

    def clean( self ):

        """
        Remove the output file (xml) from Mreps to keep only the 'set' file.
        """

        cmd = "rm -f %s" % ( self.MrepsOutFileName )
        log = os.system( cmd )
        if log != 0:
            print "*** Error in pyRepet.parser.Parser.MrepsParser.clean"
            sys.exit(1)
