import os
import sys
import time
import MySQLdb

from pyRepet.sql.RepetDBMySQL import RepetDB


#------------------------------------------------------------------------------


class Job:
    """
    Job informations to launch a command on a cluster.
    """
    
    def __init__( self, tablename="", jobid="", groupid="", queue="", \
                  command="", launcherFile="", node="", lResources="mem_free=1G" ):
        """
        Constructor.
        
        @param tablename: table name to record the jobs
        @type tablename: string
        
        @param jobid: a job identifier
        @type jobid: string
        
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        
        @param queue: queue name of the batch job manager
        @type queue: string
        
        @param command: command launched
        @type command: string
        
        @param node: cluster node name where the execution takes place
        @type node: string
        
        @param lResources: resources (memory, time...) but need to conform to SGE/PBS syntax !
        @type lResources: list of strings
        """
        
        self.tablename = tablename
        self.jobid = jobid
        self.groupid = groupid
        self.queue = queue
        self.command = command
        self.launcher = launcherFile
        self.node = node
        self.lResources = lResources


#------------------------------------------------------------------------------


class RepetJob( RepetDB ):
    """
    Connector to manage jobs launched in parallel, their features being recorded
    in a MySQL database.
    """
    
    
    def create_job_table( self, tablename ):
        """
        Create a job table.
        @param tablename: new table name
        @type tablename: string
        """
        self.remove_if_exist( tablename )
        qry = "CREATE TABLE %s (jobid VARCHAR(255), groupid VARCHAR(255), command TEXT, launcher VARCHAR(1024), queue VARCHAR(255), status VARCHAR(255), time DATETIME, node VARCHAR(255))" % ( tablename )
        self.execute( qry )
        self.update_info_table( tablename, "job table" )
        sqlCmd = "CREATE INDEX igroupid ON %s ( groupid(10) )" % ( tablename )
        self.execute( sqlCmd )
        
        
    def record_job( self, job ):
        """
        Record a job.
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        """
        self.remove_job( job )
        qry = "INSERT INTO %s" % ( job.tablename )
        qry += " VALUES (\"%s\"," % ( job.jobid )
        qry += " \"%s\"," % ( job.groupid )
        qry += " \"%s\"," % ( job.command.replace("\"","\'") )
        qry += " \"%s\"," % ( job.launcher )
        qry += " \"%s\"," % ( job.queue )
        qry += " \"waiting\","
        qry += " current_timestamp(),"
        qry += " \"?\");"
        self.execute( qry )
        while self.get_job_status(job) == "unknown":
            pass
        
        
    def remove_job( self, job ):
        """
        Remove a job from the job table.
        @param tablename: table name to record the jobs
        @type tablename: string
        @param jobid: a job identifier
        @type jobid: string
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        @param queue: queue name of the batch job manager
        @type queue: string
        @param command: command launched
        @type command: string
        """
        qry = "DELETE FROM %s WHERE jobid='%s' and groupid='%s' and queue='%s';" \
              % ( job.tablename, job.jobid, job.groupid, job.queue )
        self.execute( qry )
        
        
    def change_job_status( self, job, status, method ):
        """
        Change a job status.
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        @param status: the new status (waiting,finished,error)
        @type status: string
        """
        qry = "UPDATE %s SET status='%s', node='%s' WHERE jobid='%s' and groupid='%s' and queue='%s';" \
              % ( job.tablename, status, job.node, job.jobid, job.groupid, job.queue )
        self.execute( qry )
        
        
    def get_job_status( self, job ):
        """
        Get a job status.
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        """
        qry = "SELECT status from %s WHERE jobid='%s' and groupid='%s' and queue='%s';" \
              % ( job.tablename, job.jobid, job.groupid, job.queue )
        self.execute( qry )
        res = self.fetchall()
        if res == None or len(res) != 1:
            return "unknown"
        return res[0][0]
    
    
    def getCountStatus( self, tablename, groupid, status ):
        """
        Get the number of jobs belonging to the desired groupid with the desired status.
        @param tablename: table name to record the jobs
        @type tablename: string
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        @param status: job status (waiting, running, finished, error)
        @type status: string
        @return: number of jobs belonging to the desired groupid with the desired status
        @rtype: integer
        """
        qry = "SELECT count(jobid) FROM %s WHERE groupid='%s' AND status='%s';" \
              % ( tablename, groupid, status )
        self.execute( qry )
        res = self.fetchall()
        if len(res) == 0:
            return 0
        else:
            return int( res[0][0] )
        
        
    def clean_job_group( self, tablename, groupid ):
        """
        Clean all job from a job group.
        @param tablename: table name to record the jobs
        @type tablename: string
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        """
        if self.exist( tablename ):
            qry = "DELETE FROM %s WHERE groupid='%s';" % ( tablename, groupid )
            self.execute( qry )
            
            
    def has_unfinished_job( self, tablename, groupid ):
        """
        Check if there is unfinished job from a job group.
        @param tablename: table name to record the jobs
        @type tablename: string
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        """
        if not self.exist( tablename ):
            return False
        qry = "SELECT * FROM %s WHERE groupid='%s' and status!='finished';" \
              %( tablename, groupid )
        self.execute( qry )
        res = self.fetchall()
        if len(res) == 0:
            return False
        return True
    
    
    def wait_job_group( self, tablename, groupid, checkInterval=5, maxRelaunch=3, exitIfTooManyErrors=True ):
        """
        Wait job finished status from a job group.
        @note: job are re-launched if error (max. 3 times)
        @param tablename: table name to record the jobs
        @type tablename: string
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        @param checkInterval: time laps in seconds between two checks
        @type checkInterval: integer (default=5)
        @param maxRelaunch: max nb of times a job in error is relaunch before exiting
        @type maxRelaunch: integer (default=3)
        @param exitIfTooManyErrors: exit if a job is still in error above maxRelaunch
        @type exitIfTooManyErrors: boolean (default=True)
        """
        
        dJob2Err = {}
        
        # retrieve the total number of jobs belonging to the desired groupid
        qry = "SELECT count(jobid) FROM %s WHERE groupid='%s';" % ( tablename, groupid )
        self.execute( qry )
        totalNbJobs = int( self.fetchall()[0][0] )
        
        while True:
	    time.sleep( checkInterval )

            # retrieve the finished jobs and stop if all jobs are finished
            nbFinishedJobs = self.getCountStatus( tablename, groupid, "finished" )
            if nbFinishedJobs == totalNbJobs:
                break

            # retrieve the jobs in error and relaunch them if they are in error (max. 3 times)
            qry = "SELECT * FROM %s WHERE groupid='%s' and status ='error';" \
                  % ( tablename, groupid )
            self.execute( qry )
            lJobsInError = self.fetchall()   # 0:jobid, 1:groupid, 2:cmd, 3:launcher, 4:queue, 5:status, 6:time, 7:node
            for job in lJobsInError:
                if not dJob2Err.has_key( job[0] ):
                    dJob2Err[ job[0] ] = 1
                if dJob2Err[ job[0] ] < maxRelaunch:
                    cmd = "%s in error, re-submitting (%i)" % ( job[0], dJob2Err[ job[0] ] )
                    print cmd; sys.stdout.flush()
                    newJob = Job( tablename, job[0], groupid, job[4], job[2], job[3] )
                    self.submit_job( newJob )
                    dJob2Err[ job[0] ] += 1
                else:
                    dJob2Err[ job[0] ] += 1
                    cmd = "%s in permanent error (>%i)" % ( job[0], maxRelaunch )
                    cmd += "\ngroupid = %s" % ( groupid )
                    cmd += "\nnb of jobs = %i" % ( totalNbJobs )
                    cmd += "\nnb of finished jobs = %i" % ( self.getCountStatus( tablename, groupid, "finished" ) )
                    cmd += "\nnb of waiting jobs = %i" % ( self.getCountStatus( tablename, groupid, "waiting" ) )
                    cmd += "\nnb of running jobs = %i" % ( self.getCountStatus( tablename, groupid, "running" ) )
                    cmd += "\nnb of jobs in error = %i" % ( self.getCountStatus( tablename, groupid, "error" ) )
                    print cmd; sys.stdout.flush()
                    if exitIfTooManyErrors:
                        self.clean_job_group( tablename, groupid )
                        sys.exit(1)
                    else:
                        checkInterval = 60
                        
                        
    def submit_job( self, job, maxNbWaitingJobs=50, checkInterval=30 ):
        """
        Submit a job to a queue and record it in job table.
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        @param maxNbWaitingJobs: max nb of waiting jobs before submitting a new one
        @type maxNbWaitingJobs: integer (default=50)
        @param checkInterval: time laps in seconds between two checks
        @type checkInterval: integer (default=30)
        """
        
        if os.environ.get( "REPET_QUEUE" ) not in [ "SGE", "PBS" ]:
            print "ERROR: environment variable REPET_QUEUE should be 'SGE' or 'PBS'"
            return 1
        
        if not self.exist( job.tablename ):
            self.create_job_table( job.tablename )
            
        if self.get_job_status(job) in [ "waiting", "running", "finished" ]:
            string = "WARNING: job '%s' already submitted" % ( job.jobid )
            print string
            return
        
        self.record_job( job )
        
        cmd = "qsub"
        cmd += " -q %s" % ( job.queue )
        cmd += " -V"
        cmd += " -N %s" % ( job.jobid )
        
        if os.environ.get( "REPET_QUEUE" ) == "SGE":
            cmd += " -cwd"
            if job.lResources != []:
                cmd += " -l \""
                for resource in job.lResources:
                    cmd += resource
                cmd += "\""
                
        elif os.environ.get( "REPET_QUEUE" ) == "PBS":
            if job.lResources != []:
                cmd += " -l \""
                for resource in job.lResources:
                    cmd += resource.replace("mem_free","mem")
                cmd += "\""
                
        cmd += " <<EOF\n"
        cmd += "%s\n" % ( job.launcher )
        cmd += "EOF\n"
        
        sql = "SELECT COUNT(jobid) FROM %s WHERE groupid='%s' and status='waiting';" % ( job.tablename, job.groupid )
        self.execute( sql )
        res = self.fetchall()
        while int(res[0][0]) >= maxNbWaitingJobs:
            time.sleep( checkInterval )
            self.execute( sql )
            res = self.fetchall()
            
        return os.system( cmd )
