# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


"""
Adaptator for a L{Path<pyRepetUnit.commons.coord.Path.Path>} table.
"""

import os
from pyRepetUnit.commons.coord.Range import Range
from pyRepetUnit.commons.coord.Range import getIdx
from pyRepetUnit.commons.sql.TablePathAdaptator import TablePathAdaptator

class TableBinPathAdaptator( TablePathAdaptator ):
    """
    Bin Adaptator for a L{Path<pyRepetUnit.commons.coord.Path.Path>} table.
    """
    def __init__(self, db, table=""):
        """
        constructor

        @param db: L{RepetDB<RepetDB>} instance
        @param _table: table name
        """
        TablePathAdaptator.__init__(self, db, table)
        self._table_idx= self._table+"_idx"
        self.create_bin()

    def create_bin( self, verbose=0 ):
        """
        Create a bin table for fast access
        """
        if not self._iDb.exist(self._table):
            return
        
        if verbose > 0:
            print "creating",self._table_idx,"for fast access"
        

        self._iDb.dropTable(self._table_idx)
        sql_cmd="CREATE TABLE %s ( path int unsigned, idx int unsigned, contig varchar(255), min int, max int, strand int unsigned)"% (self._table_idx)
        self._iDb.execute(sql_cmd)

        sql_cmd="CREATE INDEX id ON %s ( path );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
        sql_cmd="CREATE INDEX ibin ON %s ( idx );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
        sql_cmd="CREATE INDEX icontig ON %s ( contig );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
        sql_cmd="CREATE INDEX imin ON %s ( min );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
        sql_cmd="CREATE INDEX imax ON %s ( max );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
        sql_cmd="CREATE INDEX istrand ON %s ( strand );"% (self._table_idx)
        self._iDb.execute(sql_cmd)
                
        tmp_file="/tmp/"+self._table+".tmp"+str(os.getpid())
        out=open(tmp_file,"w")

        _table_range=self._table+"_range"
        self._iDb.dropTable(_table_range)
        self._iDb.path2path_range(self._table)
        table=TablePathAdaptator(self._iDb,_table_range)
        if not table.isEmpty():
            contigs=table.getQueryList()
            for c in contigs:
                lpath=table.getPathListFromQuery(c)
                for i in lpath:
                    idx=i.range_query.findIdx()
                    max=i.range_query.getMax()
                    min=i.range_query.getMin()
                    strand=i.range_query.isOnDirectStrand()
                    out.write("%d\t%d\t%s\t%d\t%d\t%d\n"%(i.id,idx,i.range_query.seqname,min,max,strand))
            out.close();

            sql_cmd="LOAD DATA LOCAL INFILE '%s' INTO TABLE %s FIELDS ESCAPED BY '' "%\
                     (tmp_file,self._table_idx)
            self._iDb.execute(sql_cmd)
            self._iDb.update_info_table(self._table_idx,self._table+" bin indexes")
            os.system("rm -f "+tmp_file)


        
    def insAPath(self,path,delayed=False):
        """
        Insert a L{Path<pyRepet.coord.Path.Path>} instance

        @param path: a path object of L{Path<pyRepet.coord.Path.Path>} instance       
        @param delayed: a boolean indicating if the insert must be delayed        
        """
        TablePathAdaptator.insert(self,path,delayed)
        path.range_query.seqname=path.range_query.seqname.replace("\\","\\\\")
        path.range_subject.seqname=path.range_subject.seqname.replace("\\","\\\\")
        idx=path.range_query.findIdx()
        max=path.range_query.getMax()
        min=path.range_query.getMin()
        strand=path.range_query.isOnDirectStrand()
        if delayed:
            sql_cmd='INSERT DELAYED INTO %s VALUES (%d,%d,"%s",%d,%d,%d)'\
                 %(self._table_idx,\
                   path.id,\
                   idx,\
                   path.range_query.seqname,\
                   min,\
                   max,\
                   strand)
        else:
            sql_cmd='INSERT INTO %s VALUES (%d,%d,"%s",%d,%d,%d)'\
                 %(self._table_idx,\
                   path.id,\
                   idx,\
                   path.range_query.seqname,\
                   min,\
                   max,\
                   strand)
            
        self._iDb.execute(sql_cmd)
        
    def getPathList_from_qcoord(self,contig,start,end):
        """
        Return a L{Path<pyRepet.coord.Path.Path>} instance list overlaping
        a given region using the bin scheme

        @param contig: contig name
        @type contig: string
        @param start: start coordinate
        @type start: integer
        @param end: end coordinate
        @type end: integer
        """

        min_coord=min(start,end)
        max_coord=max(start,end)
        sql_cmd='select distinct path from %s where contig="%s" and ('\
                 %(self._table+"_idx",contig)
                 
        for bin_lvl in xrange(6,2,-1):
            if getIdx(start,bin_lvl) == getIdx(end,bin_lvl):
                idx=getIdx(start,bin_lvl)
                sql_cmd+='idx=%d'%(idx)
            else:
                idx1=getIdx(min_coord,bin_lvl)
                idx2=getIdx(max_coord,bin_lvl)
                sql_cmd+='idx between %d and %d'%(idx1,idx2)
            if bin_lvl>3:
                sql_cmd+=" or "
                
        sql_cmd+=") and min<=%d and max>=%d;"%(max_coord,min_coord)

        
        self._iDb.execute(sql_cmd)
        res=self._iDb.fetchall()
        lnum=[]
        for i in res:
            lnum.append(int(i[0]))
        lpath=self.getPathListFromIdList(lnum)
        return lpath